import os
from pathlib import Path
from tempfile import TemporaryDirectory
from django.test import TestCase
from django.conf import settings
from django_distill.renderer import copy_static_and_media_files


def null(*args, **kwargs):
    pass


class DjangoDistillStaticFilesTestSuite(TestCase):

    def test_copying_static_files(self):
        # Test default behavior
        with TemporaryDirectory() as tempdir:
            copy_static_and_media_files(tempdir, null)
            test_media_file_path = str(Path(tempdir) / 'media' / 'media-test.txt')
            self.assertTrue(os.path.exists(test_media_file_path))
            test_static_file_path = str(Path(tempdir) / 'static' / 'static-test.txt')
            self.assertTrue(os.path.exists(test_static_file_path))
        # Test admin dir is copied with SKIP_ADMIN_DIRS set to False
        settings.DISTILL_SKIP_ADMIN_DIRS = False
        with TemporaryDirectory() as tempdir:
            copy_static_and_media_files(tempdir, null)
            test_media_file_path = str(Path(tempdir) / 'media' / 'media-test.txt')
            self.assertTrue(os.path.exists(test_media_file_path))
            test_static_file_path = str(Path(tempdir) / 'static' / 'static-test.txt')
            self.assertTrue(os.path.exists(test_static_file_path))
            test_admin_file_path = str(Path(tempdir) / 'static' / 'admin' / 'admin-test.txt')
            self.assertTrue(os.path.exists(test_admin_file_path))
            test_appdir_file_path = str(Path(tempdir) / 'static' / 'appdir' / 'appdir-test.txt')
            self.assertTrue(os.path.exists(test_appdir_file_path))
        # Test static app dirs are skipped with DISTILL_SKIP_STATICFILES_DIRS populated
        settings.DISTILL_SKIP_STATICFILES_DIRS = ['appdir']
        with TemporaryDirectory() as tempdir:
            copy_static_and_media_files(tempdir, null)
            test_appdir_file_path = str(Path(tempdir) / 'static' / 'appdir' / 'appdir-test.txt')
            self.assertFalse(os.path.exists(test_appdir_file_path))



import os
import sys
import warnings
from base64 import b64decode
from binascii import hexlify


try:
    from googleapiclient import discovery
    from google.cloud import storage
except ImportError:
    name = 'django_distill.backends.google_storage'
    pipm = 'google-api-python-client google-cloud-storage'
    sys.stdout.write('{} backend requires {}:\n'.format(name, pipm))
    sys.stdout.write('$ pip install django-distill[google]{}\n\n'.format(pipm))
    raise


from django_distill.errors import DistillPublishError
from django_distill.backends import BackendBase


class GoogleCloudStorageBackend(BackendBase):
    '''
        Publisher for Google Cloud Storage. Implements the BackendBase.
    '''

    REQUIRED_OPTIONS = ('ENGINE', 'BUCKET')

    def account_username(self):
        return

    def account_container(self):
        return self.options.get('BUCKET', '')

    def authenticate(self):
        credentials_file = self.options.get('JSON_CREDENTIALS', '')
        if credentials_file:
            if not os.path.exists(credentials_file):
                err = 'Credentials file does not exist: {}'
                raise DistillPublishError(err.format(credentials_file))
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = credentials_file
        
        bucket = self.account_container()
        self.d['connection'] = storage.Client()
        self.d['bucket'] = self.d['connection'].get_bucket(bucket)

    def list_remote_files(self):
        rtn = set()
        for b in self.d['bucket'].list_blobs():
            rtn.add(b.name)
        return rtn

    def delete_remote_file(self, remote_name):
        b = self.d['bucket'].get_blob(remote_name)
        return b.delete()

    def compare_file(self, local_name, remote_name):
        b = self.d['bucket'].get_blob(remote_name)
        local_hash = self._get_local_file_hash(local_name)
        remote_hash = str(hexlify(b64decode(b.md5_hash)).decode())
        return local_hash == remote_hash

    def upload_file(self, local_name, remote_name):
        b = self.d['bucket'].blob(remote_name)
        b.upload_from_filename(local_name)
        b.make_public()
        return True

    def create_remote_dir(self, remote_dir_name):
        # not required for Google Storage buckets
        return True

    def remote_path(self, local_name):
        truncated_path = super().remote_path(local_name)
        # Replace \ for /, Google Cloud Files API docs state they handle both \ and /
        # as directory separators but really make sure we're only using / in blob names
        return truncated_path.replace('\\', '/')


backend_class = GoogleCloudStorageBackend



import os
import sys
import warnings
from urllib.parse import urlsplit, urlunsplit, quote_plus
from time import sleep
from base64 import b64decode
from binascii import hexlify


try:
    from azure.storage.blob import BlobServiceClient, ContentSettings
except ImportError:
    name = 'django_distill.backends.azure_storage'
    pipm = 'azure-storage-blob'
    sys.stdout.write('{} backend requires {}:\n'.format(name, pipm))
    sys.stdout.write('$ pip install django-distill[microsoft]{}\n\n'.format(pipm))
    raise


from django_distill.errors import DistillPublishError
from django_distill.backends import BackendBase


class AzureBlobStorateBackend(BackendBase):
    '''
        Publisher for Azure Blob Storage. Implements the BackendBase. Azure
        static websites in containers are relatively slow to make the files
        available via the public URL. To work around this uploaded files
        are cached and then verified in a loop at the end with up to
        RETRY_ATTEMPTS attempts with a delay of SLEEP_BETWEEN_RETRIES seconds
        between each attempt.
    '''

    REQUIRED_OPTIONS = ('ENGINE', 'CONNECTION_STRING')
    RETRY_ATTEMPTS = 30
    SLEEP_BETWEEN_RETRIES = 3

    def account_username(self):
        return

    def account_container(self):
        return '$web'

    def connection_string(self):
        return self.options.get('CONNECTION_STRING', '')

    def _get_blob(self, name):
        return self.d['connection'].get_blob_client(
            container=self.account_container(),
            blob=name
        )

    def _get_blob_url(self, blob):
        blob_parts = urlsplit(blob.url)
        prefix = '/{}/'.format(quote_plus(self.account_container()))
        path = blob_parts.path
        if path.startswith(prefix):
            path = path[len(prefix):]
        parts = (self.remote_url_parts.scheme, self.remote_url_parts.netloc,
                 path, None, None)
        return urlunsplit(parts)

    def authenticate(self):
        self.d['connection'] = BlobServiceClient.from_connection_string(
            conn_str=self.connection_string()
        )

    def list_remote_files(self):
        container = self.d['connection'].get_container_client(
            container=self.account_container()
        )
        rtn = set()
        for obj in container.list_blobs():
            rtn.add(obj.name)
        return rtn

    def delete_remote_file(self, remote_name):
        container = self.d['connection'].get_container_client(
            container=self.account_container()
        )
        return container.delete_blob(remote_name)

    def check_file(self, local_name, url):
        # Azure uploads are checked in bulk at the end of the uploads, do
        # nothing here
        return True

    def compare_file(self, local_name, remote_name):
        blob = self._get_blob(remote_name)
        properties = blob.get_blob_properties()
        content_settings = properties.content_settings
        content_md5 = properties.get('content_settings', {}).get('content_md5')
        if not content_md5:
            return False
        local_hash = self._get_local_file_hash(local_name)
        remote_hash = str(hexlify(bytes(content_md5)).decode())
        return local_hash == remote_hash

    def upload_file(self, local_name, remote_name):
        blob = self._get_blob(remote_name)
        mimetype = self.local_file_mimetype(local_name)
        content_settings = ContentSettings(content_type=mimetype)
        with open(local_name, 'rb') as data:
            result = blob.upload_blob(
                data, overwrite=True, content_settings=content_settings)
            if result:
                actual_url = self._get_blob_url(blob)
                self.d.setdefault('azure_uploads_to_check', []).append(
                    (local_name, remote_name, actual_url)
                )
        return result

    def _check_file(self, local_name, actual_url):
        # Azure specific patched check_file with retries to account for Azure
        # being slow
        err = ('Failed to upload local file "{}" blob to Azure container at '
               'URL "{}" not available over the public URL after {} attempts')
        local_hash = self._get_local_file_hash(local_name)
        for i in range(self.RETRY_ATTEMPTS):
            remote_hash = self._get_url_hash(actual_url)
            if not remote_hash:
                sleep(self.SLEEP_BETWEEN_RETRIES)
                continue
            if local_hash == remote_hash:
                return True
        DistillPublishError(err.format(local_name, actual_url, i + 1))

    def final_checks(self):
        # Iterate over any cached files to check and verify they have been
        # uploaded correctly.
        to_check = self.d.setdefault('azure_uploads_to_check', [])
        for (local_name, remote_name, actual_url) in to_check:
            # Verify the upload, this may require retries
            self._check_file(local_name, actual_url)
        # If we reached here no DistillPublishError was raised
        return True

    def create_remote_dir(self, remote_dir_name):
        # not required for Azure Blob Storage containers
        return True


backend_class = AzureBlobStorateBackend



from importlib import import_module
from django.test import TestCase


class DjangoDistillCommandTestSuite(TestCase):

    def test_command_imports_distill_local(self):
        import_module('django_distill.management.commands.distill-local')

    def test_command_imports_distill_publish(self):
        import_module('django_distill.management.commands.distill-publish')

    def test_command_imports_distill_test_publish(self):
        import_module('django_distill.management.commands.distill-test-publish')



import os
from shutil import rmtree
from django.core.management.base import (BaseCommand, CommandError)
from django.conf import settings
from django_distill.distill import urls_to_distill
from django_distill.renderer import (run_collectstatic, render_to_dir,
                                     copy_static_and_media_files, render_redirects)
from django_distill.errors import DistillError


class Command(BaseCommand):

    help = 'Generates a static local site using distill'

    def add_arguments(self, parser):
        parser.add_argument('output_dir', nargs='?', type=str)
        parser.add_argument('--collectstatic', dest='collectstatic', action='store_true')
        parser.add_argument('--quiet', dest='quiet', action='store_true')
        parser.add_argument('--force', dest='force', action='store_true')
        parser.add_argument('--exclude-staticfiles', dest='exclude_staticfiles', action='store_true')
        parser.add_argument('--generate-redirects', dest='generate_redirects', action='store_true')
        parser.add_argument('--parallel-render', dest='parallel_render', type=int, default=1)

    def _quiet(self, *args, **kwargs):
        pass

    def handle(self, *args, **options):
        output_dir = options.get('output_dir')
        collectstatic = options.get('collectstatic')
        quiet = options.get('quiet')
        force = options.get('force')
        exclude_staticfiles = options.get('exclude_staticfiles')
        generate_redirects = options.get('generate_redirects')
        parallel_render = options.get('parallel_render')
        if quiet:
            stdout = self._quiet
        else:
            stdout = self.stdout.write
        if not output_dir:
            output_dir = getattr(settings, 'DISTILL_DIR', None)
            if not output_dir:
                e = 'Usage: ./manage.py distill-local [directory]'
                raise CommandError(e)
        if collectstatic:
            run_collectstatic(stdout)
        if not exclude_staticfiles and not os.path.isdir(settings.STATIC_ROOT):
            e = 'Static source directory does not exist, run collectstatic'
            raise CommandError(e)
        output_dir = os.path.abspath(os.path.expanduser(output_dir))
        stdout('')
        stdout('You have requested to create a static version of')
        stdout('this site into the output path directory:')
        stdout('')
        stdout('    Source static path:  {}'.format(settings.STATIC_ROOT))
        stdout('    Distill output path: {}'.format(output_dir))
        stdout('')
        if os.path.isdir(output_dir):
            stdout('Distill output directory exists, clean up?')
            stdout('This will delete and recreate all files in the output dir')
            stdout('')
            if force:
                ans = 'yes'
            else:
                ans = input('Type \'yes\' to continue, or \'no\' to cancel: ')
            if ans.lower() == 'yes':
                stdout('Recreating output directory...')
                rmtree(output_dir)
                os.makedirs(output_dir)
            else:
                raise CommandError('Distilling site cancelled.')
        else:
            if force:
                ans = 'yes'
            else:
                ans = input('Does not exist, create it? (YES/no): ')
            if ans.lower() == 'yes':
                stdout('Creating directory...')
                os.makedirs(output_dir)
            else:
                raise CommandError('Aborting...')
        stdout('')
        stdout('Generating static site into directory: {}'.format(output_dir))
        try:
            render_to_dir(output_dir, urls_to_distill, stdout, parallel_render=parallel_render)
            if not exclude_staticfiles:
                copy_static_and_media_files(output_dir, stdout)
        except DistillError as err:
            raise CommandError(str(err)) from err
        stdout('')
        if generate_redirects:
            stdout('Generating redirects')
            render_redirects(output_dir, stdout)
            stdout('')
        stdout('Site generation complete.')



__version__ = '3.2.5'


from django import __version__ as django_version
from django_distill.errors import DistillError
from django_distill.distill import urls_to_distill
from django_distill.renderer import generate_urls


try:
    from django_distill.distill import distill_url
except ImportError:
    def distill_url(*args, **kwargs):
        err = ('Your installed version of Django ({}) does not supprt '
               'django.urls.url or django.conf.urls.url, use '
               'django_distill.distill_re_path or django_distill.distill_path')
        raise DistillError(err.format(django_version))


try:
    from django_distill.distill import distill_re_path
except ImportError:
    def distill_re_path(*args, **kwargs):
        err = ('Your installed version of Django ({}) does not supprt '
               'django.urls.re_path, please upgrade')
        raise DistillError(err.format(django_version))


try:
    from django_distill.distill import distill_path
except ImportError:
    def distill_path(*args, **kwargs):
        err = ('Your installed version of Django ({}) does not supprt '
               'django.urls.path, please upgrade')
        raise DistillError(err.format(django_version))


def distilled_urls():
    return generate_urls(urls_to_distill)



import sys
import mimetypes

try:
    import boto3
except ImportError:
    name = 'django_distill.backends.amazon_s3'
    pipm = 'boto3'
    sys.stdout.write('{} backend requires {}:\n'.format(name, pipm))
    sys.stdout.write('$ pip install django-distill[amazon]{}\n\n'.format(pipm))
    raise

from django_distill.errors import DistillPublishError
from django_distill.backends import BackendBase


class AmazonS3Backend(BackendBase):
    '''
        Publisher for Amazon S3. Implements the BackendBase.
    '''

    REQUIRED_OPTIONS = ('ENGINE', 'PUBLIC_URL', 'BUCKET')

    def _get_object(self, name):
        bucket = self.account_container()
        return self.d['connection'].head_object(Bucket=bucket, Key=name)

    def account_username(self):
        return self.options.get('ACCESS_KEY_ID', '')

    def account_container(self):
        return self.options.get('BUCKET', '')

    def authenticate(self, calling_format=None):
        access_key_id = self.account_username()
        secret_access_key = self.options.get('SECRET_ACCESS_KEY', '')
        endpoint_url = self.options.get('ENDPOINT_URL', '')
        bucket = self.account_container()
        if access_key_id and secret_access_key:
            self.d['connection'] = boto3.client('s3', aws_access_key_id=access_key_id,
                                                aws_secret_access_key=secret_access_key,
                                                endpoint_url=endpoint_url)
        else:
            self.d['connection'] = boto3.client('s3')
        self.d['bucket'] = bucket

    def list_remote_files(self):
        rtn = set()
        response = self.d['connection'].list_objects_v2(Bucket=self.d['bucket'])
        if 'Contents' in response:
            for obj in response['Contents']:
                rtn.add(obj['Key'])
        return rtn

    def delete_remote_file(self, remote_name):
        self.d['connection'].delete_object(Bucket=self.d['bucket'], Key=remote_name)

    def compare_file(self, local_name, remote_name):
        obj = self._get_object(remote_name)
        local_hash = self._get_local_file_hash(local_name)
        return local_hash == obj['ETag'][1:-1]

    def upload_file(self, local_name, remote_name):
        content_type, _ = mimetypes.guess_type(local_name)
        if content_type is None:
            content_type = 'application/octet-stream'
        extra_args = {'ContentType': content_type}
        self.d['connection'].upload_file(local_name, self.d['bucket'], remote_name, ExtraArgs=extra_args)

    def create_remote_dir(self, remote_dir_name):
        # not required for S3 buckets
        return True

backend_class = AmazonS3Backend



from django_distill.errors import DistillError


urls_to_distill = []


def _distill_url(func, *a, **k):
    distill_func = k.get('distill_func')
    if distill_func:
        del k['distill_func']
    else:
        distill_func = lambda: None
    distill_file = k.get('distill_file')
    distill_status_codes = k.get('distill_status_codes')
    if distill_file:
        del k['distill_file']
    if distill_status_codes:
        del k['distill_status_codes']
    else:
        distill_status_codes = (200,)
    name = k.get('name')
    if not name:
        raise DistillError('Distill function provided with no name')
    if not callable(distill_func):
        err = 'Distill function not callable: {}'
        raise DistillError(err.format(distill_func))
    url = func(*a, **k)
    urls_to_distill.append((url, distill_func, distill_file, distill_status_codes,
                            name, a, k))
    return url


try:
    from django.conf.urls import url
    def distill_url(*a, **k):
        return _distill_url(url, *a, **k)
except ImportError:
    try:
        from django.urls import url
        def distill_url(*a, **k):
            return _distill_url(url, *a, **k)
    except ImportError:
        pass


try:
    from django.urls import path
    def distill_path(*a, **k):
        return _distill_url(path, *a, **k)
except ImportError:
    pass


try:
    from django.urls import re_path
    def distill_re_path(*a, **k):
        return _distill_url(re_path, *a, **k)
except ImportError:
    pass



from django.test import TestCase
from django.conf import settings
import django_distill


class DjangoDistillInterfaceTestSuite(TestCase):

    def test_import(self):
        assert hasattr(django_distill, 'distill_url')
        assert callable(django_distill.distill_url)
        if settings.HAS_RE_PATH:
            assert hasattr(django_distill, 'distill_re_path')
            assert callable(django_distill.distill_re_path)
        if settings.HAS_PATH:
            assert hasattr(django_distill, 'distill_path')
            assert callable(django_distill.distill_path)



from django.http import HttpResponse
from django_distill import distill_path


app_name = 'sub-urls'


def test_url_in_deep_namespace_view(request):
    return HttpResponse(b'test', content_type='application/octet-stream')


def test_no_param_func():
    return None


urlpatterns = [

    distill_path('sub-url-in-sub-namespace',
        test_url_in_deep_namespace_view,
        name='test_url_in_namespace',
        distill_func=test_no_param_func,
        distill_file='test_url_in_sub_namespace'),

]



import os
import sys
import tempfile
from datetime import datetime, timedelta
from unittest.mock import patch
from django.test import TestCase, override_settings
from django.conf import settings
from django.contrib.flatpages.models import FlatPage
from django.apps import apps as django_apps
from django.utils import timezone
from django.utils.translation import activate as activate_lang
from django_distill.distill import urls_to_distill
from django_distill.renderer import DistillRender, render_to_dir, render_single_file, get_renderer
from django_distill.errors import DistillError
from django_distill import distilled_urls


class CustomRender(DistillRender):
    pass


class DjangoDistillRendererTestSuite(TestCase):

    def setUp(self):
        self.renderer = DistillRender(urls_to_distill)
        # Create a few test flatpages
        Site = django_apps.get_model('sites.Site')
        current_site = Site.objects.get_current()
        page1 = FlatPage()
        page1.url = '/flat/page1.html'
        page1.title = 'flatpage1'
        page1.content = 'flatpage1'
        page1.template_name = 'flatpage.html'
        page1.save()
        page1.sites.add(current_site)
        page2 = FlatPage()
        page2.url = '/flat/page2.html'
        page2.title = 'flatpage2'
        page2.content = 'flatpage2'
        page2.template_name = 'flatpage.html'
        page2.save()
        page2.sites.add(current_site)

    def _get_view(self, name):
        for u in urls_to_distill:
            if u[4] == name:
                return u
        return False

    def _skip(self, what):
        sys.stdout.write('Missing {}, skipping test... '.format(what))
        sys.stdout.flush()

    def test_is_str(self):
        self.assertTrue(self.renderer._is_str('a'))
        self.assertFalse(self.renderer._is_str(None))
        self.assertFalse(self.renderer._is_str(1))
        self.assertFalse(self.renderer._is_str([]))
        self.assertFalse(self.renderer._is_str(()))
        self.assertFalse(self.renderer._is_str({}))
        self.assertFalse(self.renderer._is_str({'a':'a'}))
        self.assertFalse(self.renderer._is_str(object()))

    def test_get_uri_values(self):
        test = ()
        check = self.renderer.get_uri_values(lambda: test, None)
        self.assertEqual(check, (None,))
        test = ('a',)
        check = self.renderer.get_uri_values(lambda: test, None)
        self.assertEqual(check, test)
        test = (('a',),)
        check = self.renderer.get_uri_values(lambda: test, None)
        self.assertEqual(check, test)
        test = []
        check = self.renderer.get_uri_values(lambda: test, None)
        self.assertEqual(check, (None,))
        test = ['a']
        check = self.renderer.get_uri_values(lambda: test, None)
        self.assertEqual(check, test)
        test = [['a']]
        check = self.renderer.get_uri_values(lambda: test, None)
        self.assertEqual(check, test)
        for invalid in ('a', 1, b'a', {'s'}, {'a':'a'}, object()):
            with self.assertRaises(DistillError):
                self.renderer.get_uri_values(lambda: invalid, None)

    def test_re_path_no_param(self):
        if not settings.HAS_RE_PATH:
            self._skip('django.urls.re_path')
            return
        view = self._get_view('re_path-no-param')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        if not param_set:
            param_set = ()
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        self.assertEqual(uri, '/re_path/')
        render = self.renderer.render_view(uri, status_codes, param_set, args)
        self.assertEqual(render.content, b'test')

    def test_re_path_positional_param(self):
        if not settings.HAS_RE_PATH:
            self._skip('django.urls.re_path')
            return
        view = self._get_view('re_path-positional-param')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_sets = self.renderer.get_uri_values(view_func, view_name)
        for param_set in param_sets:
            param_set = (param_set,)
            first_value = param_set[0]
            uri = self.renderer.generate_uri(view_url, view_name, param_set)
            self.assertEqual(uri, f'/re_path/{first_value}')
            render = self.renderer.render_view(uri, status_codes, param_set, args)
            self.assertEqual(render.content, b'test' + first_value.encode())

    def test_re_path_named_param(self):
        if not settings.HAS_RE_PATH:
            self._skip('django.urls.re_path')
            return
        view = self._get_view('re_path-named-param')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        self.assertEqual(uri, '/re_path/test')
        render = self.renderer.render_view(uri, status_codes, param_set, args)
        self.assertEqual(render.content, b'testtest')

    def test_re_broken(self):
        if not settings.HAS_RE_PATH:
            self._skip('django.urls.re_path')
            return
        view = self._get_view('re_path-broken')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        if not param_set:
            param_set = ()
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        self.assertEqual(uri, '/re_path/broken')
        with self.assertRaises(DistillError):
            self.renderer.render_view(uri, status_codes, param_set, args)

    def test_path_no_param(self):
        if not settings.HAS_PATH:
            self._skip('django.urls.path')
            return
        view = self._get_view('path-no-param')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        if not param_set:
            param_set = ()
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        self.assertEqual(uri, '/path/')
        render = self.renderer.render_view(uri, status_codes, param_set, args)
        self.assertEqual(render.content, b'test')

    def test_path_positional_param(self):
        if not settings.HAS_PATH:
            self._skip('django.urls.path')
            return
        view = self._get_view('path-positional-param')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_sets = self.renderer.get_uri_values(view_func, view_name)
        for param_set in param_sets:
            param_set = (param_set,)
            first_value = param_set[0]
            uri = self.renderer.generate_uri(view_url, view_name, param_set)
            self.assertEqual(uri, f'/path/{first_value}')
            render = self.renderer.render_view(uri, status_codes, param_set, args)
            self.assertEqual(render.content, b'test' + first_value.encode())

    def test_path_named_param(self):
        if not settings.HAS_PATH:
            self._skip('django.urls.path')
            return
        view = self._get_view('path-named-param')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        self.assertEqual(uri, '/path/test')
        render = self.renderer.render_view(uri, status_codes, param_set, args)
        self.assertEqual(render.content, b'testtest')

    def test_path_broken(self):
        if not settings.HAS_PATH:
            self._skip('django.urls.path')
            return
        view = self._get_view('path-broken')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        if not param_set:
            param_set = ()
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        self.assertEqual(uri, '/path/broken')
        with self.assertRaises(DistillError):
            self.renderer.render_view(uri, status_codes, param_set, args)

    def test_render_paths(self):
        def _blackhole(_):
            pass
        expected_files = (
            ('test',),
            ('re_path', '12345'),
            ('re_path', 'test'),
            ('re_path', 'x', '12345.html'),
            ('re_path', 'x', 'test.html'),
        )
        with tempfile.TemporaryDirectory() as tmpdirname:
            with self.assertRaises(DistillError):
                render_to_dir(tmpdirname, urls_to_distill, _blackhole)
            written_files = []
            for (root, dirs, files) in os.walk(tmpdirname):
                for f in files:
                    filepath = os.path.join(root, f)
                    written_files.append(filepath)
            for expected_file in expected_files:
                filepath = os.path.join(tmpdirname, *expected_file)
                self.assertIn(filepath, written_files)

    @patch.object(CustomRender, "render_view", side_effect=CustomRender.render_view, autospec=True)
    @override_settings(DISTILL_RENDERER="tests.test_renderer.CustomRender")
    def test_render_paths_custom_renderer(self, render_view_spy):
        def _blackhole(_):
            pass
        expected_files = (
            ('test',),
            ('re_path', '12345'),
            ('re_path', 'test'),
        )
        with tempfile.TemporaryDirectory() as tmpdirname:
            with self.assertRaises(DistillError):
                render_to_dir(tmpdirname, urls_to_distill, _blackhole)
            written_files = []
            for (root, dirs, files) in os.walk(tmpdirname):
                for f in files:
                    filepath = os.path.join(root, f)
                    written_files.append(filepath)
            for expected_file in expected_files:
                filepath = os.path.join(tmpdirname, *expected_file)
                self.assertIn(filepath, written_files)
        #self.assertEqual(render_view_spy.call_count, 34)

    def test_sessions_are_ignored(self):
        if settings.HAS_PATH:
            view = self._get_view('path-ignore-sessions')
            assert view
            view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
            param_set = self.renderer.get_uri_values(view_func, view_name)[0]
            if not param_set:
                param_set = ()
            uri = self.renderer.generate_uri(view_url, view_name, param_set)
            self.assertEqual(uri, '/path/ignore-sessions')
            render = self.renderer.render_view(uri, status_codes, param_set, args)
            self.assertEqual(render.content, b'test')
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        if settings.HAS_RE_PATH:
            view = self._get_view('re_path-ignore-sessions')
            assert view
            view_url, iew_func, file_name, status_codes, view_name, args, kwargs = view
            param_set = self.renderer.get_uri_values(view_func, view_name)[0]
            if not param_set:
                param_set = ()
            uri = self.renderer.generate_uri(view_url, view_name, param_set)
            self.assertEqual(uri, '/re_path/ignore-sessions')
            render = self.renderer.render_view(uri, status_codes, param_set, args)
            self.assertEqual(render.content, b'test')

    def test_custom_status_codes(self):
        if settings.HAS_PATH:
            view = self._get_view('path-404')
            assert view
            view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
            param_set = self.renderer.get_uri_values(view_func, view_name)[0]
            if not param_set:
                param_set = ()
            uri = self.renderer.generate_uri(view_url, view_name, param_set)
            self.assertEqual(uri, '/path/404')
            render = self.renderer.render_view(uri, status_codes, param_set, args)
            self.assertEqual(render.content, b'404')
            self.assertEqual(render.status_code, 404)
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        if settings.HAS_RE_PATH:
            view = self._get_view('re_path-404')
            assert view
            view_url, iew_func, file_name, status_codes, view_name, args, kwargs = view
            param_set = self.renderer.get_uri_values(view_func, view_name)[0]
            if not param_set:
                param_set = ()
            uri = self.renderer.generate_uri(view_url, view_name, param_set)
            self.assertEqual(uri, '/re_path/404')
            render = self.renderer.render_view(uri, status_codes, param_set, args)
            self.assertEqual(render.content, b'404')
            self.assertEqual(render.status_code, 404)

    def test_contrib_flatpages(self):
        if settings.HAS_PATH:
            view = self._get_view('path-flatpage')
            assert view
            view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
            param_set = self.renderer.get_uri_values(view_func, view_name)
            for param in param_set:
                page_url = param['url']
                uri = self.renderer.generate_uri(view_url, view_name, param)
                self.assertEqual(uri, f'/path/flatpage{page_url}')
                render = self.renderer.render_view(uri, status_codes, param, args)
                flatpage = FlatPage.objects.get(url=page_url)
                expected = f'<title>{flatpage.title}</title><body>{flatpage.content}</body>\n'
                self.assertEqual(render.content, expected.encode())
                self.assertEqual(render.status_code, 200)
        if settings.HAS_RE_PATH:
            view = self._get_view('re_path-flatpage')
            assert view
            view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
            param_set = self.renderer.get_uri_values(view_func, view_name)
            for param in param_set:
                page_url = param['url']
                uri = self.renderer.generate_uri(view_url, view_name, param)
                self.assertEqual(uri, f'/re_path/flatpage{page_url}')
                render = self.renderer.render_view(uri, status_codes, param, args)
                flatpage = FlatPage.objects.get(url=page_url)
                expected = f'<title>{flatpage.title}</title><body>{flatpage.content}</body>\n'
                self.assertEqual(render.content, expected.encode())
                self.assertEqual(render.status_code, 200)

    def test_contrib_sitemaps(self):
        view = self._get_view('path-sitemap')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = ()
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        self.assertEqual(uri, '/path/test-sitemap')
        render = self.renderer.render_view(uri, status_codes, param_set, args)
        expected_content = (
            b'<?xml version="1.0" encoding="UTF-8"?>\n'
            b'<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9" '
            b'xmlns:xhtml="http://www.w3.org/1999/xhtml">\n'
            b'<url><loc>http://example.com/path/test-sitemap</loc>'
            b'<changefreq>daily</changefreq>'
            b'<priority>0.5</priority></url>\n</urlset>\n'
        )
        self.assertEqual(render.content, expected_content)
        self.assertEqual(render.status_code, 200)

    def test_render_single_file(self):
        expected_files = (
            ('path', '12345'),
            ('path', 'test'),
        )
        with tempfile.TemporaryDirectory() as tmpdirname:
            render_single_file(tmpdirname, 'path-positional-param', 12345)
            render_single_file(tmpdirname, 'path-named-param', param='test')
            written_files = []
            for (root, dirs, files) in os.walk(tmpdirname):
                for f in files:
                    filepath = os.path.join(root, f)
                    written_files.append(filepath)
            for expected_file in expected_files:
                filepath = os.path.join(tmpdirname, *expected_file)
                self.assertIn(filepath, written_files)

    def test_i18n(self):
        if not settings.USE_I18N:
            self._skip('settings.USE_I18N')
            return
        settings.DISTILL_LANGUAGES = [
            'en',
            'fr',
            'de',
        ]
        expected = {}
        for lang_code, lang_name in settings.DISTILL_LANGUAGES:
            expected[lang_code] = f'/{lang_code}/path/i18n/sub-url-with-i18n-prefix'
        view = self._get_view('test-url-i18n')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        if not param_set:
            param_set = ()
        for lang_code, path in expected.items():
            activate_lang(lang_code)
            uri = self.renderer.generate_uri(view_url, view_name, param_set)
            self.assertEqual(uri, path)
            render = self.renderer.render_view(uri, status_codes, param_set, args)
            self.assertEqual(render.content, b'test')
        # Render the test URLs and confirm the expected language URI prefixes are present
        def _blackhole(_):
            pass
        expected_files = (
            ('test',),
            ('re_path', '12345'),
            ('re_path', 'test'),
            ('re_path', 'x', '12345.html'),
            ('re_path', 'x', 'test.html'),
            ('en', 'path', 'i18n', 'sub-url-with-i18n-prefix'),
            ('fr', 'path', 'i18n', 'sub-url-with-i18n-prefix'),
            ('de', 'path', 'i18n', 'sub-url-with-i18n-prefix'),
        )
        with tempfile.TemporaryDirectory() as tmpdirname:
            with self.assertRaises(DistillError):
                render_to_dir(tmpdirname, urls_to_distill, _blackhole, parallel_render=8)
            written_files = []
            for (root, dirs, files) in os.walk(tmpdirname):
                for f in files:
                    filepath = os.path.join(root, f)
                    written_files.append(filepath)
            for expected_file in expected_files:
                filepath = os.path.join(tmpdirname, *expected_file)
                self.assertIn(filepath, written_files)
        settings.DISTILL_LANGUAGES = []

    def test_kwargs(self):
        if not settings.HAS_PATH:
            self._skip('django.urls.path')
            return
        view = self._get_view('test-kwargs')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        if not param_set:
            param_set = ()
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        self.assertEqual(uri, '/path/kwargs')
        render = self.renderer.render_view(uri, status_codes, param_set, args, kwargs)
        self.assertEqual(render.content, b'test')

    def test_humanize(self):
        if not settings.HAS_PATH:
            self._skip('django.urls.path')
            return
        view = self._get_view('test-humanize')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        if not param_set:
            param_set = ()
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        self.assertEqual(uri, '/path/humanize')
        now = timezone.now()
        one_hour_ago = now - timedelta(hours=1)
        nineteen_hours_ago = now - timedelta(hours=19)
        render = self.renderer.render_view(uri, status_codes, param_set, args, kwargs)
        content = render.content
        expected = b'\n'.join([
            b'',
            b'<ul>',
            b'<li>test</li>',
            b'<li>one hour ago naturaltime: an hour ago</li>',
            b'<li>nineteen hours ago naturaltime: 19\xc2\xa0hours ago</li>',
            b'</ul>',
            b'',
        ])
        self.assertEqual(render.content, expected)

    def test_request_has_resolver_match(self):
        view = self._get_view('test-has-resolver-match')
        assert view
        view_url, view_func, file_name, status_codes, view_name, args, kwargs = view
        param_set = self.renderer.get_uri_values(view_func, view_name)[0]
        if not param_set:
            param_set = ()
        uri = self.renderer.generate_uri(view_url, view_name, param_set)
        render = self.renderer.render_view(uri, status_codes, param_set, args, kwargs)
        self.assertEqual(render.content, b"test_request_has_resolver_match")

    def test_parallel_rendering(self):
        def _blackhole(_):
            pass
        expected_files = (
            ('test',),
            ('re_path', '12345'),
            ('re_path', 'test'),
            ('re_path', 'x', '12345.html'),
            ('re_path', 'x', 'test.html'),
        )
        with tempfile.TemporaryDirectory() as tmpdirname:
            with self.assertRaises(DistillError):
                render_to_dir(tmpdirname, urls_to_distill, _blackhole, parallel_render=8)
            written_files = []
            for (root, dirs, files) in os.walk(tmpdirname):
                for f in files:
                    filepath = os.path.join(root, f)
                    written_files.append(filepath)
            for expected_file in expected_files:
                filepath = os.path.join(tmpdirname, *expected_file)
                self.assertIn(filepath, written_files)

    def test_generate_urls(self):
        urls = distilled_urls()
        generated_urls = []
        for url, file_name in urls:
            generated_urls.append(url)
        expected_urls = (
            '/path/namespace1/sub-url-in-namespace',
            '/path/namespace1/path/sub-namespace/sub-url-in-sub-namespace',
            '/path/no-namespace/sub-url-in-no-namespace',
            '/en/path/i18n/sub-url-with-i18n-prefix',
            '/re_path/',
            '/re_path-no-func/',
            '/re_path/12345',
            '/re_path/67890',
            '/re_path/x/12345',
            '/re_path/x/67890',
            '/re_path/test',
            '/re_path/x/test',
            '/re_path/broken',
            '/re_path/ignore-sessions',
            '/re_path/404',
            '/re_path/flatpage/flat/page1.html',
            '/re_path/flatpage/flat/page2.html',
            '/path/',
            '/path-no-func/',
            '/path/12345',
            '/path/67890',
            '/path/x/12345',
            '/path/x/67890',
            '/path/test',
            '/path/x/test',
            '/path/broken',
            '/path/ignore-sessions',
            '/path/404',
            '/path/flatpage/flat/page1.html',
            '/path/flatpage/flat/page2.html',
            '/path/test-sitemap',
            '/path/kwargs',
            '/path/humanize',
            '/path/has-resolver-match'
        )
        self.assertEqual(sorted(generated_urls), sorted(expected_urls))



import os
import sys
import warnings
import mimetypes
from hashlib import md5
from binascii import hexlify
import requests
from urllib.parse import urlsplit, urlunsplit
from django_distill.errors import DistillPublishError
from django_distill.renderer import filter_dirs


class BackendBase(object):
    '''
        Generic base class for all backends, mostly an interface / template.
    '''

    REQUIRED_OPTIONS = ()

    def __init__(self, source_dir, options):
        if not source_dir.endswith(os.sep):
            source_dir += os.sep
        self.source_dir = source_dir
        self.options = options
        self.local_files = set()
        self.local_dirs = set()
        self.remote_files = set()
        self.remote_url_parts = urlsplit(options.get('PUBLIC_URL', ''))
        self.d = {}
        self._validate_options()

    def _validate_options(self):
        for o in self.REQUIRED_OPTIONS:
            if o not in self.options:
                e = 'Missing required settings value for this backend: {}'
                raise DistillPublishError(e.format(o))

    def index_local_files(self):
        for root, dirs, files in os.walk(self.source_dir):
            dirs[:] = filter_dirs(dirs)
            for d in dirs:
                self.local_dirs.add(os.path.join(root, d))
            for f in files:
                self.local_files.add(os.path.join(root, f))

    def _get_local_file_hash(self, file_path, digest_func=md5, chunk=1048576):
        # md5 is used by Amazon S3, Rackspace Cloud Files and Google Storage
        if not self._file_exists(file_path):
            return None
        digest = digest_func()
        with open(file_path, 'rb') as f:
            while True:
                data = f.read(chunk)
                if not data:
                    break
                digest.update(data)
        return digest.hexdigest()

    def _get_url_hash(self, url, digest_func=md5, chunk=1024):
        # CDN cache buster
        url += '?' + hexlify(os.urandom(16)).decode('utf-8')
        request = requests.get(url, stream=True)
        if request.status_code == 404:
            return False
        digest = digest_func()
        for block in request.iter_content(chunk_size=chunk):
            if block:
                digest.update(block)
        return digest.hexdigest()

    def _file_exists(self, file_path):
        return os.path.isfile(file_path)

    def local_file_mimetype(self, local_name):
        try:
            return mimetypes.guess_type(local_name)[0]
        except Exception:
            return 'application/octet-stream'

    def remote_url(self, local_name):
        if local_name[:len(self.source_dir)] != self.source_dir:
            raise DistillPublishError('File {} is not in source dir {}'.format(
                local_name, self.source_dir))
        truncated = local_name[len(self.source_dir):]
        remote_file = '/'.join(truncated.split(os.sep))
        remote_uri = self.remote_url_parts.path + remote_file
        return urlunsplit((self.remote_url_parts.scheme,
                           self.remote_url_parts.netloc, remote_uri, '', ''))

    def list_local_dirs(self):
        return self.local_dirs

    def list_local_files(self):
        return self.local_files

    def check_file(self, local_name, url):
        if not self._file_exists(local_name):
            raise DistillPublishError('File does not exist: {}'.format(
                local_name))
        local_hash = self._get_local_file_hash(local_name)
        remote_hash = self._get_url_hash(url)
        return local_hash == remote_hash

    def final_checks(self):
        pass

    def remote_path(self, local_name):
        return local_name[len(self.source_dir):]

    def account_username(self):
        raise NotImplementedError('account_username must be implemented')

    def account_container(self):
        raise NotImplementedError('account_container must be implemented')

    def authenticate(self):
        raise NotImplementedError('authenticate must be implemented')

    def list_remote_files(self):
        raise NotImplementedError('list_remote_files must be implemented')

    def delete_remote_file(self, remote_name):
        raise NotImplementedError('delete_remote_file must be implemented')

    def compare_file(self, local_name, remote_name):
        raise NotImplementedError('compare_file must be implemented')

    def upload_file(self, local_name, remote_name):
        raise NotImplementedError('upload_file must be implemented')

    def create_remote_dir(self, remote_dir_name):
        raise NotImplementedError('create_remote_dir must be implemented')


def get_backend(engine):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        try:
            backend = __import__(engine, globals(), locals(),
                                 ['backend_class'])
        except ImportError as e:
            err = 'Failed to import backend engine: {} ({})\n'
            sys.stderr.write(err.format(engine, e))
            raise
    module = getattr(backend, 'backend_class')
    if not module:
        raise ImportError('Backend engine has no backend_class attribute')
    return module



import errno
import inspect
import logging
import os
import types
from shutil import copy2
from concurrent.futures import ThreadPoolExecutor
from django.utils.translation import activate as activate_lang
from django.conf import settings, global_settings
from django.urls import include as include_urls, get_resolver
from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed
from django.utils.module_loading import import_string
from django.test import RequestFactory
from django.test.client import ClientHandler
from django.urls import reverse, ResolverMatch
from django.urls.exceptions import NoReverseMatch
from django.core.management import call_command
from django_distill.errors import DistillError


logger = logging.getLogger(__name__)
urlconf = get_resolver()


def iter_resolved_urls(url_patterns, namespace_path=[]):
    url_patterns_resolved = []
    for entry in url_patterns:
        if hasattr(entry, 'url_patterns'):
            if getattr(entry, 'namespace', None) is not None:
                url_patterns_resolved += iter_resolved_urls(
                    entry.url_patterns, namespace_path + [entry.namespace])
            else:
                url_patterns_resolved += iter_resolved_urls(
                    entry.url_patterns, namespace_path)
        else:
            url_patterns_resolved.append((namespace_path, entry))
    return url_patterns_resolved


def load_namespace_map():
    namespace_map = {}
    for (namespaces, url) in iter_resolved_urls(urlconf.url_patterns):
        if namespaces:
            nspath = ':'.join(namespaces)
            if url in namespace_map:
                raise DistillError(f'Ambiguous namespace for URL "{url}" in namespace '
                                   f'"{nspath}", Distill does not support the same Django '
                                   f'app being include()ed more than once in the same '
                                   f'project')
            else:
                namespace_map[url] = nspath
    return namespace_map


class DistillHandler(ClientHandler):
    '''
        Overload ClientHandler's resolve_request(...) to return the already known
        and pre-resolved view method. Also overwrite any session handling middleware
        with the dummy session handler.
    '''

    def __init__(self, *a, **k):
        self.view_func = lambda x: x
        self.view_uri_args = ()
        self.view_uri_kwargs = {}
        self.view_args = ()
        k['enforce_csrf_checks'] = False
        super().__init__(*a, *k)

    def set_view(self, view_func, view_uri_args, view_uri_kwargs, view_args):
        self.view_func = view_func
        self.view_uri_args = view_uri_args
        self.view_uri_kwargs = view_uri_kwargs
        self.view_args = view_args

    def resolve_request(self, request):
        for arg in self.view_args:
            self.view_uri_kwargs.update(**arg)
        request.resolver_match = ResolverMatch(
            self.view_func,
            self.view_uri_args,
            self.view_uri_kwargs,
        )
        return request.resolver_match

    def load_middleware(self, is_async=False):
        '''
            Replaces the standard BaseHandler.load_middleware(). This method is
            identical apart from not trapping all exceptions. We actually want the
            real Python exceptions to be raised here.
        '''
        self._view_middleware = []
        self._template_response_middleware = []
        self._exception_middleware = []
        get_response = self._get_response_async if is_async else self._get_response
        handler = get_response
        handler_is_async = is_async
        for middleware_path in reversed(settings.MIDDLEWARE):
            middleware = import_string(middleware_path)
            middleware_can_sync = getattr(middleware, 'sync_capable', True)
            middleware_can_async = getattr(middleware, 'async_capable', False)
            if not middleware_can_sync and not middleware_can_async:
                raise RuntimeError(
                    'Middleware %s must have at least one of '
                    'sync_capable/async_capable set to True.' % middleware_path
                )
            elif not handler_is_async and middleware_can_sync:
                middleware_is_async = False
            else:
                middleware_is_async = middleware_can_async
            try:
                # Adapt handler, if needed.
                adapted_handler = self.adapt_method_mode(
                    middleware_is_async,
                    handler,
                    handler_is_async,
                    debug=settings.DEBUG,
                    name='middleware %s' % middleware_path,
                )
                mw_instance = middleware(adapted_handler)
            except MiddlewareNotUsed as exc:
                if settings.DEBUG:
                    if str(exc):
                        logger.debug('MiddlewareNotUsed(%r): %s', middleware_path, exc)
                    else:
                        logger.debug('MiddlewareNotUsed: %r', middleware_path)
                continue
            else:
                handler = adapted_handler
            if mw_instance is None:
                raise ImproperlyConfigured(
                    'Middleware factory %s returned None.' % middleware_path
                )
            if hasattr(mw_instance, 'process_view'):
                self._view_middleware.insert(
                    0,
                    self.adapt_method_mode(is_async, mw_instance.process_view),
                )
            if hasattr(mw_instance, 'process_template_response'):
                self._template_response_middleware.append(
                    self.adapt_method_mode(
                        is_async, mw_instance.process_template_response
                    ),
                )
            if hasattr(mw_instance, 'process_exception'):
                self._exception_middleware.append(
                    self.adapt_method_mode(False, mw_instance.process_exception),
                )
            handler = mw_instance
            handler_is_async = middleware_is_async
        handler = self.adapt_method_mode(is_async, handler, handler_is_async)
        self._middleware_chain = handler


class DistillRender(object):
    '''
        Renders a complete static site from all urls registered with
        distill_url() and then copies over all static media.
    '''

    def __init__(self, urls_to_distill, parallel_render=1):
        self.urls_to_distill = urls_to_distill
        self.parallel_render = parallel_render
        self.namespace_map = load_namespace_map()
        # set allowed hosts to '*', static rendering shouldn't care about the hostname
        settings.ALLOWED_HOSTS = ['*']

    def render_file(self, view_name, status_codes, view_args, view_kwargs):
        view_details = []
        for params in self.urls_to_distill:
            if view_name == params[4]:
                view_details = params
                break
        if not view_details:
            raise DistillError(f'No view exists with the name: {view_name}')
        url, distill_func, file_name, status_codes, view_name, a, k = view_details
        if view_kwargs:
            uri = self.generate_uri(url, view_name, view_kwargs)
            args = view_kwargs
        else:
            uri = self.generate_uri(url, view_name, view_args)
            args = view_args
        render = self.render_view(uri, status_codes, args, a, k)
        file_name = self._get_filename(file_name, uri, args)
        return uri, file_name, render

    def render_all_urls(self, do_render=True):

        def _render(item):
            rtn = []
            for lang in self.get_langs():
                activate_lang(lang)
                url, view_name, param_set, status_codes, file_name_base, a, k, do_render = item
                uri = self.generate_uri(url, view_name, param_set)
                if do_render:
                    render = self.render_view(uri, status_codes, param_set, a, k)
                else:
                    render = None
                file_name = self._get_filename(file_name_base, uri, param_set)
                rtn.append((uri, file_name, render))
            return rtn

        to_render = []
        for url, distill_func, file_name_base, status_codes, view_name, a, k in self.urls_to_distill:
            for param_set in self.get_uri_values(distill_func, view_name):
                if not param_set:
                    param_set = ()
                elif self._is_str(param_set):
                    param_set = (param_set,)
                to_render.append((url, view_name, param_set, status_codes, file_name_base, a, k, do_render))
        with ThreadPoolExecutor(max_workers=self.parallel_render) as executor:
            results = executor.map(_render, to_render)
            for i18n_result in results:
                for uri, file_name, render in i18n_result:
                    yield uri, file_name, render

    def render(self, view_name=None, status_codes=None, view_args=None, view_kwargs=None):
        if view_name:
            if not status_codes:
                status_codes = (200,)
            if not view_args:
                view_args = ()
            if not view_kwargs:
                view_kwargs = {}
            return self.render_file(view_name, status_codes, view_args, view_kwargs)
        else:
            return self.render_all_urls()

    def urls(self):
        for (uri, file_name, _) in self.render_all_urls(do_render=False):
            yield uri, file_name

    def get_langs(self):
        langs = []
        LANGUAGE_CODE = str(getattr(settings, 'LANGUAGE_CODE', 'en'))
        GLOBAL_LANGUAGES = list(getattr(global_settings, 'LANGUAGES', []))
        try:
            LANGUAGES = list(getattr(settings, 'LANGUAGES', []))
        except (ValueError, TypeError, AttributeError):
            LANGUAGES = []
        try:
            DISTILL_LANGUAGES = list(getattr(settings, 'DISTILL_LANGUAGES', []))
        except (ValueError, TypeError, AttributeError) as e:
            DISTILL_LANGUAGES = []
        if LANGUAGES != GLOBAL_LANGUAGES:
            for lang_code, lang_name in LANGUAGES:
                langs.append(lang_code)
        if LANGUAGE_CODE not in DISTILL_LANGUAGES and LANGUAGE_CODE not in langs:
            langs.append(LANGUAGE_CODE)
        for lang in DISTILL_LANGUAGES:
            langs.append(lang)
        return langs

    def _get_filename(self, file_name, uri, param_set):
        if file_name is not None:
            if isinstance(param_set, dict):
                return file_name.format(**param_set)
            else:
                return file_name.format(*param_set)
        elif uri.endswith('/'):
            # rewrite URIs ending with a slash to ../index.html
            if uri.startswith('/'):
                uri = uri[1:]
            return uri + 'index.html'
        else:
            return None

    def _is_str(self, s):
        return isinstance(s, str)

    def get_uri_values(self, func, view_name):
        fullargspec = inspect.getfullargspec(func)
        try:
            if 'view_name' in fullargspec.args:
                v = func(view_name)
            else:
                v = func()
        except Exception as e:
            raise DistillError('Failed to call distill function: {}'.format(e))
        if not v:
            return (None,)
        elif isinstance(v, (list, tuple)):
            return v
        elif isinstance(v, types.GeneratorType):
            return list(v)
        else:
            err = 'Distill function returned an invalid type: {}'
            raise DistillError(err.format(type(v)))

    def generate_uri(self, url, view_name, param_set):
        namespace = self.namespace_map.get(url, '')
        view_name_ns = namespace + ':' + view_name if namespace else view_name
        if isinstance(param_set, (list, tuple)):
            try:
                uri = reverse(view_name, args=param_set)
            except NoReverseMatch:
                uri = reverse(view_name_ns, args=param_set)
        elif isinstance(param_set, dict):
            try:
                uri = reverse(view_name, kwargs=param_set)
            except NoReverseMatch:
                uri = reverse(view_name_ns, kwargs=param_set)
        else:
            err = 'Distill function returned an invalid type: {}'
            raise DistillError(err.format(type(param_set)))
        return uri

    def render_view(self, uri, status_codes, param_set, args, kwargs={}):
        view_path, view_func = None, None
        try:
            view_path, view_func = args[0], args[1]
        except IndexError:
            try:
                view_path, view_func = kwargs['route'], kwargs['view']
            except KeyError:
                pass
        if view_path is None or view_func is None:
            raise DistillError(f'Invalid view arguments, args:{args}, kwargs:{kwargs}')
        # Default status_codes to (200,) if they are invalid or not set
        if not isinstance(status_codes, (tuple, list)):
            status_codes = (200,)
        for status_code in status_codes:
            if not isinstance(status_code, int):
                status_codes = (200,)
                break
        view_args = args[2:] if len(args) > 2 else ()
        request_factory = RequestFactory()
        request = request_factory.get(uri)
        handler = DistillHandler()
        handler.load_middleware()
        if isinstance(param_set, dict):
            a, k = (), param_set
        else:
            a, k = param_set, {}
        try:
            handler.set_view(view_func, a, k, view_args)
            response = handler.get_response(request)
        except Exception as err:
            e = 'Failed to render view "{}": {}'.format(uri, err)
            raise DistillError(e) from err
        if response.status_code not in status_codes:
            err = 'View returned an invalid status code: {} (expected one of {})'
            raise DistillError(err.format(response.status_code, status_codes))
        return response


def copy_static(dir_from, dir_to):
    # we need to ignore some static dirs such as 'admin' so this is a
    # little more complex than a straight shutil.copytree()
    if not dir_from.endswith(os.sep):
        dir_from = dir_from + os.sep
    if not dir_to.endswith(os.sep):
        dir_to = dir_to + os.sep
    for root, dirs, files in os.walk(dir_from):
        dirs[:] = filter_dirs(dirs)
        for f in files:
            from_path = os.path.join(root, f)
            base_path = from_path[len(dir_from):]
            to_path = os.path.join(dir_to, base_path)
            to_path_dir = os.path.dirname(to_path)
            if not os.path.isdir(to_path_dir):
                os.makedirs(to_path_dir)
            copy2(from_path, to_path)
            yield from_path, to_path


def copy_static_and_media_files(output_dir, stdout):
    static_url = str(settings.STATIC_URL)
    static_root = str(settings.STATIC_ROOT)
    static_url = static_url[1:] if static_url.startswith('/') else static_url
    static_output_dir = os.path.join(output_dir, static_url)
    for file_from, file_to in copy_static(static_root, static_output_dir):
        stdout('Copying static: {} -> {}'.format(file_from, file_to))
    media_url = str(settings.MEDIA_URL)
    media_root = str(settings.MEDIA_ROOT)
    if media_root:
        media_url = media_url[1:] if media_url.startswith('/') else media_url
        media_output_dir = os.path.join(output_dir, media_url)
        for file_from, file_to in copy_static(media_root, media_output_dir):
            stdout('Copying media: {} -> {}'.format(file_from, file_to))
    return True


def run_collectstatic(stdout):
    stdout('Distill is running collectstatic...')
    call_command('collectstatic', '--noinput')
    stdout('')
    stdout('collectstatic complete, continuing...')


def filter_dirs(dirs):
    DISTILL_SKIP_ADMIN_DIRS = bool(getattr(settings, 'DISTILL_SKIP_ADMIN_DIRS', True))
    _ignore_dirs = []
    if DISTILL_SKIP_ADMIN_DIRS:
        _ignore_dirs.append('admin')
        _ignore_dirs.append('grappelli')
    try:
        DISTILL_SKIP_STATICFILES_DIRS = list(getattr(settings, 'DISTILL_SKIP_STATICFILES_DIRS', []))
    except (ValueError, TypeError):
        DISTILL_SKIP_STATICFILES_DIRS = []
    for d in DISTILL_SKIP_STATICFILES_DIRS:
        if isinstance(d, str) and os.sep not in d:
            _ignore_dirs.append(d)
    return [d for d in dirs if d not in _ignore_dirs]


def load_urls(stdout=None):
    if stdout:
        stdout('Loading site URLs')
    for url in urlconf.url_patterns:
        include_urls(url)


def get_filepath(output_dir, file_name, page_uri):
    if file_name:
        local_uri = file_name
        full_path = os.path.join(output_dir, file_name)
    else:
        local_uri = page_uri
        if page_uri.startswith('/'):
            page_uri = page_uri[1:]
        page_path = page_uri.replace('/', os.sep)
        full_path = os.path.join(output_dir, page_path)
    return full_path, local_uri


def write_file(full_path, content):
    try:
        dirname = os.path.dirname(full_path)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        with open(full_path, 'wb') as f:
            f.write(content)
    except IOError as e:
        if e.errno == errno.EISDIR:
            err = ('Output path: {} is a directory! Try adding a '
                   '"distill_file" arg to your distill_url()')
            raise DistillError(err.format(full_path))
        else:
            raise


def get_renderer(urls_to_distill, parallel_render=1):
    import_path = getattr(settings, 'DISTILL_RENDERER', None)
    if import_path:
        render_cls = import_string(import_path)
    else:
        render_cls = DistillRender
    return render_cls(urls_to_distill, parallel_render)


def render_to_dir(output_dir, urls_to_distill, stdout, parallel_render=1):
    load_urls(stdout)
    renderer = get_renderer(urls_to_distill, parallel_render)
    for page_uri, file_name, http_response in renderer.render():
        full_path, local_uri = get_filepath(output_dir, file_name, page_uri)
        content = http_response.content
        mime = http_response.get('Content-Type')
        renamed = ' (renamed from "{}")'.format(page_uri) if file_name else ''
        msg = 'Rendering page: {} -> {} ["{}", {} bytes] {}'
        stdout(msg.format(local_uri, full_path, mime, len(content), renamed))
        write_file(full_path, content)
    return True


def render_single_file(output_dir, view_name, *args, **kwargs):
    from django_distill.distill import urls_to_distill
    status_codes = None
    if 'status_codes' in kwargs:
        status_codes = kwargs['status_codes']
        del kwargs['status_codes']
    view_name = str(view_name)
    if not status_codes:
        status_codes = (200,)
    load_urls()
    renderer = get_renderer(urls_to_distill)
    page_uri, file_name, http_response = renderer.render(
        view_name, status_codes, args, kwargs)
    full_path, local_uri = get_filepath(output_dir, file_name, page_uri)
    content = http_response.content
    write_file(full_path, content)
    return True


def generate_urls(urls_to_distill):
    load_urls()
    renderer = get_renderer(urls_to_distill)
    return renderer.urls()


def render_static_redirect(destination_url):
    redir = []
    redir.append(f'<!DOCTYPE html>')
    redir.append(f'<html>')
    redir.append(f'<head>')
    redir.append(f'<meta charset="UTF-8">')
    redir.append(f'<meta http-equiv="refresh" content="0;URL={destination_url}" />')
    redir.append(f'<title>Redirecting to {destination_url}</title>')
    redir.append(f'<meta name="robots" content="noindex" />')
    redir.append(f'</head>')
    redir.append(f'<body>')
    redir.append(f'<h1>Redirecting to <a href="{destination_url}">{destination_url}</a></h1>')
    redir.append(f'<p>If you are not automatically redirected please click <a href="{destination_url}">this link</a></p>')
    redir.append(f'</html>')
    return '\n'.join(redir).encode()


def render_redirects(output_dir, stdout):
    from django.contrib.redirects.models import Redirect
    for redirect in Redirect.objects.all():
        redirect_path = redirect.old_path.lstrip('/')
        if redirect_path.lower().endswith('.html'):
            redirect_file = redirect_path
        else:
            redirect_file = os.path.join(redirect_path, 'index.html')
        full_path, local_uri = get_filepath(output_dir, redirect_file, redirect_file)
        content = render_static_redirect(redirect.new_path)
        msg = 'Rendering redirect: {} -> {}'
        stdout(msg.format(local_uri, redirect.new_path))
        write_file(full_path, content)
    return True



from pathlib import Path


BASE_DIR = Path(__file__).resolve().parent.parent


try:
    from django.urls import path
    HAS_PATH = True
except ImportError:
    HAS_PATH = False


try:
    from django.urls import re_path
    HAS_RE_PATH = True
except ImportError:
    HAS_RE_PATH = False


SECRET_KEY = 'test'


ROOT_URLCONF = 'tests.urls'


MIDDLEWARE = ['django.contrib.sessions.middleware.SessionMiddleware']


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'test.sqlite3',
    }
}


INSTALLED_APPS = [
    'django.contrib.sites',
    'django.contrib.flatpages',
    'django.contrib.sessions',
    'django.contrib.sitemaps',
    'django.contrib.humanize',
    'django.contrib.redirects',
    'django_distill',
    'tests',
]


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [BASE_DIR / 'tests' / 'templates'],
        'APP_DIRS': True,
    },
]


SITE_ID = 1


STATIC_URL = '/static/'
STATIC_ROOT = BASE_DIR / 'tests' / 'static'
MEDIA_URL = '/media/'
MEDIA_ROOT = BASE_DIR / 'tests' / 'media'


LANGUAGE_CODE = 'en'
USE_I18N = True



from concurrent.futures import ThreadPoolExecutor
from django_distill.errors import DistillPublishError


def publish_dir(backend, stdout, verify=True, parallel_publish=1, ignore_remote_content=False):
    stdout('Authenticating')
    backend.authenticate()
    stdout('Getting file indexes')
    remote_files = set() if ignore_remote_content else backend.list_remote_files()
    local_files = backend.list_local_files()
    to_upload = set()
    to_delete = set()
    local_files_r = set()
    # check local files to upload
    for f in local_files:
        remote_f = backend.remote_path(f)
        local_files_r.add(remote_f)
        if remote_f not in remote_files:
            # local file not present remotely, upload it
            to_upload.add(f)
        else:
            # file is present remotely, check its hash
            if not backend.compare_file(f, remote_f):
                stdout(f'File stale (hash different): {remote_f}')
                to_upload.add(f)
            else:
                stdout(f'File fresh: {remote_f}')
    # check for remote files to delete
    for f in remote_files:
        if f not in local_files_r:
            to_delete.add(f)
    with ThreadPoolExecutor(max_workers=parallel_publish) as executor:
        # upload any new or changed files
        executor.map(lambda f: _publish_file(backend, f, verify, stdout), to_upload)
        # Call any final checks that may be needed by the backend
        stdout('Final checks')
        backend.final_checks()
        # delete any orphan files
        executor.map(lambda f: _delete_file(backend, f, stdout), to_delete)


def _publish_file(backend, f, verify, stdout):
    remote_f = backend.remote_path(f)
    stdout(f'Publishing: {f} -> {remote_f}')
    backend.upload_file(f, remote_f)
    if verify:
        url = backend.remote_url(f)
        stdout(f'Verifying: {url}')
        if not backend.check_file(f, url):
            err = f'Remote file {url} failed hash check'
            raise DistillPublishError(err)


def _delete_file(backend, f, stdout):
    stdout(f'Deleting remote: {f}')
    backend.delete_remote_file(f)



import os
import tempfile
from django.conf import settings
from django.core.management.base import (BaseCommand, CommandError)
from django_distill.backends import get_backend
from django_distill.distill import urls_to_distill
from django_distill.errors import DistillError
from django_distill.renderer import (run_collectstatic, render_to_dir,
                                     copy_static_and_media_files, render_redirects)
from django_distill.publisher import publish_dir


class Command(BaseCommand):

    help = 'Distills a site into a temporary local directory then publishes it'

    def add_arguments(self, parser):
        parser.add_argument('publish_target_name', nargs='?', type=str)
        parser.add_argument('--collectstatic', dest='collectstatic',
                            action='store_true')
        parser.add_argument('--quiet', dest='quiet', action='store_true')
        parser.add_argument('--force', dest='force', action='store_true')
        parser.add_argument('--exclude-staticfiles', dest='exclude_staticfiles',
                    action='store_true')
        parser.add_argument('--skip-verify', dest='skip_verify', action='store_true')
        parser.add_argument('--ignore-remote-content', dest='ignore_remote_content', action='store_true')
        parser.add_argument('--parallel-publish', dest='parallel_publish', type=int, default=1)
        parser.add_argument('--generate-redirects', dest='generate_redirects', action='store_true')
        parser.add_argument('--parallel-render', dest='parallel_render', type=int, default=1)

    def _quiet(self, *args, **kwargs):
        pass

    def handle(self, *args, **options):
        publish_target_name = options.get('publish_target_name')
        if not publish_target_name:
            publish_target_name = 'default'
        publish_targets = getattr(settings, 'DISTILL_PUBLISH', {})
        publish_target = publish_targets.get(publish_target_name)
        if type(publish_target) != dict:
            e = 'Invalid publish target name: "{}"'.format(publish_target_name)
            e += ', check your settings.DISTILL_PUBLISH values'
            raise CommandError(e)
        publish_engine = publish_target.get('ENGINE')
        if not publish_engine:
            e = 'Publish target {} has no ENGINE'.format(publish_target_name)
            raise CommandError(e)
        collectstatic = options.get('collectstatic')
        exclude_staticfiles = options.get('exclude_staticfiles')
        skip_verify = options.get('skip_verify', False)
        parallel_publish = options.get('parallel_publish')
        ignore_remote_content = options.get('ignore_remote_content', False)
        quiet = options.get('quiet')
        force = options.get('force')
        generate_redirects = options.get('generate_redirects')
        parallel_render = options.get('parallel_render')
        if quiet:
            stdout = self._quiet
        else:
            stdout = self.stdout.write
        with tempfile.TemporaryDirectory() as output_dir:
            if not output_dir.endswith(os.sep):
                output_dir += os.sep
            backend_class = get_backend(publish_engine)
            backend = backend_class(output_dir, publish_target)
            username = backend.account_username()
            container = backend.account_container()
            stdout('')
            stdout('You have requested to distill and publish this site')
            stdout('to the following target:')
            stdout('')
            stdout('    Settings name: {}'.format(publish_target_name))
            stdout('    Engine:        {}'.format(publish_engine))
            stdout('    Username:      {}'.format(username))
            stdout('    Container:     {}'.format(container))
            stdout('')
            if collectstatic:
                run_collectstatic(stdout)
            if not exclude_staticfiles and not os.path.isdir(settings.STATIC_ROOT):
                e = 'Static source directory does not exist, run collectstatic'
                raise CommandError(e)
            if force:
                ans = 'yes'
            else:
                ans = input('Type \'yes\' to continue, or \'no\' to cancel: ')
            if ans.lower() == 'yes':
                pass
            else:
                raise CommandError('Publishing site cancelled.')
            self.stdout.write('')
            msg = 'Generating static site into directory: {}'
            stdout(msg.format(output_dir))
            try:
                render_to_dir(output_dir, urls_to_distill, stdout, parallel_render=parallel_render)
                if not exclude_staticfiles:
                    copy_static_and_media_files(output_dir, stdout)
            except DistillError as err:
                raise CommandError(str(err)) from err
            stdout('')
            if generate_redirects:
                stdout('Generating redirects')
                render_redirects(output_dir, stdout)
                stdout('')
            stdout('Publishing site')
            backend.index_local_files()
            publish_dir(backend, stdout, not skip_verify, parallel_publish, ignore_remote_content)
        stdout('')
        stdout('Site generation and publishing complete.')



from django.http import HttpResponse
from django_distill import distill_path


app_name = 'i18n'


def test_url_i18n_view(request):
    return HttpResponse(b'test', content_type='application/octet-stream')


def test_no_param_func():
    return None


urlpatterns = [

    distill_path('sub-url-with-i18n-prefix',
        test_url_i18n_view,
        name='test-url-i18n',
        distill_func=test_no_param_func),

]



from datetime import timedelta
from django.conf import settings
from django.http import HttpResponse
from django.urls import include, path, reverse
from django.conf.urls.i18n import i18n_patterns
from django.shortcuts import render
from django.utils import timezone
from django.contrib.flatpages.views import flatpage as flatpage_view
from django.contrib.sitemaps import Sitemap
from django.contrib.sitemaps.views import sitemap
from django.apps import apps as django_apps
from django_distill import distill_path, distill_re_path


class TestStaticViewSitemap(Sitemap):

    priority = 0.5
    changefreq = 'daily'

    def items(self):
        return ['path-sitemap']

    def location(self, item):
        return reverse(item)


sitemap_dict = {
    'static': TestStaticViewSitemap,
}


def test_no_param_view(request):
    return HttpResponse(b'test', content_type='application/octet-stream')


def test_positional_param_view(request, param):
    if not isinstance(param, str):
        param = str(param)
    return HttpResponse(b'test' + param.encode(),
                        content_type='application/octet-stream')


def test_named_param_view(request, param=None):
    if not isinstance(param, str):
        param = str(param)
    return HttpResponse(b'test' + param.encode(),
                        content_type='application/octet-stream')


def test_session_view(request):
    request.session['test'] = 'test'
    return HttpResponse(b'test', content_type='application/octet-stream')


def test_broken_view(request):
    # Trigger a normal Python exception when rendering
    a = 1 / 0


def test_http404_view(request):
    response = HttpResponse(b'404', content_type='application/octet-stream')
    response.status_code = 404
    return response


def test_humanize_view(request):
    now = timezone.now()
    one_hour_ago = now - timedelta(hours=1)
    nineteen_hours_ago = now - timedelta(hours=19)
    return render(request, 'humanize.html', {
        'now': now,
        'one_hour_ago': one_hour_ago,
        'nineteen_hours_ago': nineteen_hours_ago
    })


def test_no_param_func():
    return None


def test_positional_param_func():
    return ('12345', '67890')


def test_named_param_func():
    return [{'param': 'test'}]


def test_flatpages_func():
    Site = django_apps.get_model('sites.Site')
    current_site = Site.objects.get_current()
    flatpages = current_site.flatpage_set.filter(registration_required=False)
    for flatpage in flatpages:
        yield {'url': flatpage.url}


def test_request_has_resolver_match(request):
    return HttpResponse(request.resolver_match.func.__name__)


urlpatterns = [

    path('path/namespace1/',
        include('tests.namespaced_urls', namespace='test_namespace')),
    path('path/no-namespace/',
        include('tests.no_namespaced_urls')),

]


urlpatterns += i18n_patterns(
    path('path/i18n/', include('tests.i18n_urls')),
)


if settings.HAS_RE_PATH:
    urlpatterns += [

        distill_re_path(r'^re_path/$',
            test_no_param_view,
            name='re_path-no-param',
            distill_func=test_no_param_func,
            distill_file='test'),
        distill_re_path(r'^re_path-no-func/$',
            test_no_param_view,
            name='re_path-no-param-no-func',
            distill_file='test'),
        distill_re_path(r'^re_path/([\d]+)$',
            test_positional_param_view,
            name='re_path-positional-param',
            distill_func=test_positional_param_func),
        distill_re_path(r'^re_path/x/([\d]+)$',
            test_positional_param_view,
            name='re_path-positional-param-custom',
            distill_func=test_positional_param_func,
            distill_file="re_path/x/{}.html"),
        distill_re_path(r'^re_path/(?P<param>[\w]+)$',
            test_named_param_view,
            name='re_path-named-param',
            distill_func=test_named_param_func),
        distill_re_path(r'^re_path/x/(?P<param>[\w]+)$',
            test_named_param_view,
            name='re_path-named-param-custom',
            distill_func=test_named_param_func,
            distill_file="re_path/x/{param}.html"),
        distill_re_path(r'^re_path/broken$',
            test_broken_view,
            name='re_path-broken',
            distill_func=test_no_param_func),
        distill_re_path(r'^re_path/ignore-sessions$',
            test_session_view,
            name='re_path-ignore-sessions',
            distill_func=test_no_param_func),
        distill_re_path(r'^re_path/404$',
            test_http404_view,
            name='re_path-404',
            distill_status_codes=(404,),
            distill_func=test_no_param_func),
        distill_re_path(r'^re_path/flatpage(?P<url>.+)$',
            flatpage_view,
            name='re_path-flatpage',
            distill_func=test_flatpages_func),

    ]


if settings.HAS_PATH:
    urlpatterns += [

        distill_path('path/',
            test_no_param_view,
            name='path-no-param',
            distill_func=test_no_param_func,
            distill_file='test'),
        distill_path('path-no-func/',
            test_no_param_view,
            name='path-no-param-no-func',
            distill_file='test'),
        distill_path('path/<int>',
            test_positional_param_view,
            name='path-positional-param',
            distill_func=test_positional_param_func),
        distill_path('path/x/<int>',
            test_positional_param_view,
            name='path-positional-param-custom',
            distill_func=test_positional_param_func,
            distill_file="path/x/{}.html"),
        distill_path('path/<str:param>',
            test_named_param_view,
            name='path-named-param',
            distill_func=test_named_param_func),
        distill_path('path/x/<str:param>',
            test_named_param_view,
            name='path-named-param-custom',
            distill_func=test_named_param_func,
            distill_file="path/x/{param}.html"),
        distill_path('path/broken',
            test_broken_view,
            name='path-broken',
            distill_func=test_no_param_func),
        distill_path('path/ignore-sessions',
            test_session_view,
            name='path-ignore-sessions',
            distill_func=test_no_param_func),
        distill_path('path/404',
            test_http404_view,
            name='path-404',
            distill_status_codes=(404,),
            distill_func=test_no_param_func),
        distill_path('path/flatpage<path:url>',
            flatpage_view,
            name='path-flatpage',
            distill_func=test_flatpages_func),
        distill_path('path/test-sitemap',
            sitemap,
            {'sitemaps': sitemap_dict},
            name='path-sitemap'
        ),
        distill_path(route='path/kwargs',
            view=test_no_param_view,
            name='test-kwargs'),
        distill_path('path/humanize',
            test_humanize_view,
            name='test-humanize'),
        distill_path(
            "path/has-resolver-match",
            test_request_has_resolver_match,
            name="test-has-resolver-match",
        ),

    ]



import os
import sys
from setuptools import setup, find_packages


version = '3.2.5'


with open('README.md', 'rt') as f:
    long_description = f.read()


with open('requirements.txt', 'rt') as f:
    requirements = tuple(f.read().split())


setup(
    name = 'django-distill',
    version = version,
    url = 'https://github.com/meeb/django-distill',
    author = 'https://github.com/meeb',
    author_email = 'meeb@meeb.org',
    description = 'Static site renderer and publisher for Django.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    license = 'MIT',
    include_package_data = True,
    install_requires = requirements,
    extras_require = {
        'amazon': ['boto3'],
        'google': ['google-api-python-client', 'google-cloud-storage'],
        'microsoft': ['azure-storage-blob'],
    },
    packages = find_packages(),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords = ['django', 'distill', 'static', 'website', 'jamstack', 's3',
                'amazon s3', 'aws', 'amazon', 'google', 'microsoft',
                'google cloud', 'google cloud storage', 'azure',
                'azure storage', 'azure blob storage'],
)



from django.http import HttpResponse
from django_distill import distill_path
from django.urls import path, include


app_name = 'namespaced-urls'


def test_url_in_namespace_view(request):
    return HttpResponse(b'test', content_type='application/octet-stream')


def test_no_param_func():
    return None


urlpatterns = [

    distill_path('sub-url-in-namespace',
        test_url_in_namespace_view,
        name='test_url_in_namespace',
        distill_func=test_no_param_func,
        distill_file='test_url_in_namespace'),
    path('path/sub-namespace/',
        include('tests.namespaced_sub_urls', namespace='sub_test_namespace')),
    # Uncomment to trigger a DistillError for including the same sub-urls more than
    # once in a single project which is unsupported
    #path('path/sub-namespace1/',
    #    include('tests.namespaced_sub_urls', namespace='sub_test_namespace1')),

]



import os
from pathlib import Path
from tempfile import TemporaryDirectory
from django.test import TestCase
from django.contrib.redirects.models import Redirect
from django_distill.renderer import render_static_redirect, render_redirects


def null(*args, **kwargs):
    pass


class DjangoDistillRedirectsTestSuite(TestCase):    

    def test_template(self):
        test_url = 'https://example.com/'
        test_template = render_static_redirect(test_url)
        expected_template = ('<!DOCTYPE html>\n'
                             '<html>\n'
                             '<head>\n'
                             '<meta charset="UTF-8">\n'
                             '<meta http-equiv="refresh" content="0;URL=https://example.com/" />\n'
                             '<title>Redirecting to https://example.com/</title>\n'
                             '<meta name="robots" content="noindex" />\n'
                             '</head>\n'
                             '<body>\n'
                             '<h1>Redirecting to <a href="https://example.com/">https://example.com/</a></h1>\n'
                             '<p>If you are not automatically redirected please click <a href="https://example.com/">this link</a></p>\n'
                             '</html>')
        self.assertEqual(test_template, expected_template.encode())

    def test_redirects(self):
        # Create some test redirects
        Redirect.objects.create(site_id=1, old_path='/redirect-from1/', new_path='/redirect-to1/')
        Redirect.objects.create(site_id=1, old_path='/redirect-from2/', new_path='/redirect-to2/')
        Redirect.objects.create(site_id=1, old_path='/redirect-from3/', new_path='/redirect-to3/')
        Redirect.objects.create(site_id=1, old_path='/redirect-from4/test.html', new_path='/redirect-to4/test.html')
        Redirect.objects.create(site_id=1, old_path='/redirect-from5/noslash', new_path='/redirect-to5/noslash')
        Redirect.objects.create(site_id=1, old_path='/redirect-from6/deep/redirect/path/', new_path='/redirect-to6/')
        # Render the redirect templates
        with TemporaryDirectory() as tempdir:
            render_redirects(tempdir, null)
            # Test the redirect templates exist
            for redirect in Redirect.objects.all():
                redirect_path = redirect.old_path.lstrip('/')
                if redirect_path.lower().endswith('.html'):
                    test_file_path = str(Path(tempdir) / redirect_path)
                else:
                    test_file_path = str(Path(tempdir) / redirect_path / 'index.html')
                self.assertTrue(os.path.exists(test_file_path))
                with open(test_file_path, 'rb') as f:
                    test_file_contents = f.read()
                    expected_file_contents = render_static_redirect(redirect.new_path)
                    self.assertEqual(test_file_contents, expected_file_contents)
